@@ -9,17 +9,29 @@ import (
 	"os"
 	"path/filepath"
 	"regexp"
+	"strconv"
 	"strings"
 	"testing"
 	"time"
 
-	"github.com/flosch/pongo2/v4"
+	"github.com/flosch/pongo2/v6"
 )
 
-var adminList = []string{"user2"}
+type stringerValueType int
 
-var time1 = time.Date(2014, 06, 10, 15, 30, 15, 0, time.UTC)
-var time2 = time.Date(2011, 03, 21, 8, 37, 56, 12, time.UTC)
+func (v stringerValueType) String() string {
+	return "-" + strconv.Itoa(int(v)) + ":"
+}
+
+var (
+	strPtr    = stringerValueType(1234)
+	adminList = []string{"user2"}
+)
+
+var (
+	time1 = time.Date(2014, 0o6, 10, 15, 30, 15, 0, time.UTC)
+	time2 = time.Date(2011, 0o3, 21, 8, 37, 56, 12, time.UTC)
+)
 
 type post struct {
 	Text    string
@@ -62,8 +74,7 @@ func (p *post) String() string {
  * Start setup sandbox
  */
 
-type tagSandboxDemoTag struct {
-}
+type tagSandboxDemoTag struct{}
 
 func (node *tagSandboxDemoTag) Execute(ctx *pongo2.ExecutionContext, writer pongo2.TemplateWriter) *pongo2.Error {
 	writer.WriteString("hello")
@@ -107,7 +118,7 @@ func init() {
 
 var tplContext = pongo2.Context{
 	"number": 11,
-	"simple": map[string]interface{}{
+	"simple": map[string]any{
 		"number":                   42,
 		"name":                     "john doe",
 		"included_file":            "INCLUDES.helper",
@@ -132,11 +143,13 @@ Yep!`,
 		"multiple_item_list": []int{1, 1, 2, 3, 5, 8, 13, 21, 34, 55},
 		"unsorted_int_list":  []int{192, 581, 22, 1, 249, 9999, 1828591, 8271},
 		"fixed_item_list":    [...]int{1, 2, 3, 4},
-		"misc_list":          []interface{}{"Hello", 99, 3.14, "good"},
+		"misc_list":          []any{"Hello", 99, 3.14, "good"},
 		"escape_text":        "This is \\a Test. \"Yep\". 'Yep'.",
 		"xss":                "<script>alert(\"uh oh\");</script>",
 		"time1":              time1,
 		"time2":              time2,
+		"stringer":           strPtr,
+		"stringerPtr":        &strPtr,
 		"intmap": map[int]string{
 			1: "one",
 			5: "five",
@@ -153,10 +166,10 @@ Yep!`,
 		"func_add": func(a, b int) int {
 			return a + b
 		},
-		"func_add_iface": func(a, b interface{}) interface{} {
+		"func_add_iface": func(a, b any) any {
 			return a.(int) + b.(int)
 		},
-		"func_variadic": func(msg string, args ...interface{}) string {
+		"func_variadic": func(msg string, args ...any) string {
 			return fmt.Sprintf(msg, args...)
 		},
 		"func_variadic_sum_int": func(args ...int) int {
@@ -175,8 +188,19 @@ Yep!`,
 			}
 			return pongo2.AsValue(s)
 		},
+		"func_ensure_nil": func(x any) bool {
+			return x == nil
+		},
+		"func_ensure_nil_variadic": func(args ...any) bool {
+			for _, i := range args {
+				if i != nil {
+					return false
+				}
+			}
+			return true
+		},
 	},
-	"complex": map[string]interface{}{
+	"complex": map[string]any{
 		"is_admin": isAdmin,
 		"post": post{
 			Text:    "<h2>Hello!</h2><p>Welcome to my new blog page. I'm using pongo2 which supports {{ variables }} and {% tags %}.</p>",
@@ -238,7 +262,7 @@ Yep!`,
 }
 
 func TestTemplate_Functions(t *testing.T) {
-	mydict := map[string]interface{}{
+	mydict := map[string]any{
 		"foo":    "bar",
 		"foobar": 8379,
 	}
@@ -256,7 +280,7 @@ func TestTemplate_Functions(t *testing.T) {
 			template: "{{ testFunc(mydict) }}",
 			context: pongo2.Context{
 				"mydict": mydict,
-				"testFunc": func(i interface{}) (string, error) {
+				"testFunc": func(i any) (string, error) {
 					d, err := json.Marshal(i)
 					return string(d), err
 				},
@@ -269,7 +293,7 @@ func TestTemplate_Functions(t *testing.T) {
 			template: "{{ testFunc(mydict) }}",
 			context: pongo2.Context{
 				"mydict": mydict,
-				"testFunc": func(i interface{}) (string, error) {
+				"testFunc": func(i any) (string, error) {
 					return "", errors.New("something went wrong")
 				},
 			},
@@ -281,7 +305,7 @@ func TestTemplate_Functions(t *testing.T) {
 			template: "{{ testFunc(mydict) }}",
 			context: pongo2.Context{
 				"mydict": mydict,
-				"testFunc": func(i interface{}) (string, int, error) {
+				"testFunc": func(i any) (string, int, error) {
 					return "", 0, nil
 				},
 			},
@@ -292,15 +316,26 @@ func TestTemplate_Functions(t *testing.T) {
 			name:     "InvalidArguments",
 			template: "{{ testFunc(mydict) }}",
 			context: pongo2.Context{
-				"mydict": map[string]interface{}{
+				"mydict": map[string]any{
 					"foo":    "bar",
 					"foobar": 8379,
 				},
-				"testFunc": func(i interface{}) (string, int) {
+				"testFunc": func(i any) (string, int) {
 					return "", 0
 				},
 			},
-			errorMessage: "[Error (where: execution) in <string> | Line 1 Col 4 near 'testFunc'] The second return value is not an error",
+			errorMessage: "[Error (where: execution) in <string> | Line 1 Col 4 near 'testFunc'] the second return value is not an error",
+			wantErr:      true,
+		},
+		{
+			name:     "NilToNonNilParameter",
+			template: "{{ testFunc(nil) }}",
+			context: pongo2.Context{
+				"testFunc": func(i int) int {
+					return 1
+				},
+			},
+			errorMessage: "[Error (where: execution) in <string> | Line 1 Col 4 near 'testFunc'] function input argument 0 of 'testFunc' must be of type int or *pongo2.Value (not <nil>)",
 			wantErr:      true,
 		},
 	}
@@ -335,7 +370,6 @@ func TestTemplates(t *testing.T) {
 	}
 	for idx, match := range matches {
 		t.Run(fmt.Sprintf("%03d-%s", idx+1, match), func(t *testing.T) {
-
 			t.Logf("[Template %3d] Testing '%s'", idx+1, match)
 			tpl, err := pongo2.FromFile(match)
 			if err != nil {
@@ -360,10 +394,10 @@ func TestTemplates(t *testing.T) {
 				t.Fatalf("Error on Execute('%s'): %s", match, err.Error())
 			}
 			tplOut = testTemplateFixes.fixIfNeeded(match, tplOut)
-			if bytes.Compare(testOut, tplOut) != 0 {
+			if !bytes.Equal(testOut, tplOut) {
 				t.Logf("Template (rendered) '%s': '%s'", match, tplOut)
 				errFilename := filepath.Base(fmt.Sprintf("%s.error", match))
-				err := ioutil.WriteFile(errFilename, []byte(tplOut), 0600)
+				err := ioutil.WriteFile(errFilename, []byte(tplOut), 0o600)
 				if err != nil {
 					t.Fatalf(err.Error())
 				}
@@ -375,48 +409,50 @@ func TestTemplates(t *testing.T) {
 }
 
 func TestBlockTemplates(t *testing.T) {
-	//debug = true
+	// debug = true
 
 	matches, err := filepath.Glob("./template_tests/block_render/*.tpl")
 	if err != nil {
 		t.Fatal(err)
 	}
 	for idx, match := range matches {
-		t.Logf("[BlockTemplate %3d] Testing '%s'", idx+1, match)
+		t.Run(fmt.Sprintf("%03d-%s", idx+1, match), func(t *testing.T) {
+			t.Logf("[BlockTemplate %3d] Testing '%s'", idx+1, match)
 
-		tpl, err := pongo2.FromFile(match)
-		if err != nil {
-			t.Fatalf("Error on FromFile('%s'): %s", match, err.Error())
-		}
-
-		testFilename := fmt.Sprintf("%s.out", match)
-		testOut, rerr := ioutil.ReadFile(testFilename)
-		if rerr != nil {
-			t.Fatalf("Error on ReadFile('%s'): %s", testFilename, rerr.Error())
-		}
-		tpl_out, err := tpl.ExecuteBlocks(tplContext, []string{"content", "more_content"})
-		if err != nil {
-			t.Fatalf("Error on ExecuteBlocks('%s'): %s", match, err.Error())
-		}
-
-		if _, ok := tpl_out["content"]; !ok {
-			t.Errorf("Failed: content not in tpl_out for %s", match)
-		}
-		if _, ok := tpl_out["more_content"]; !ok {
-			t.Errorf("Failed: more_content not in tpl_out for %s", match)
-		}
-		testString := string(testOut[:])
-		joinedString := strings.Join([]string{tpl_out["content"], tpl_out["more_content"]}, "")
-		if testString != joinedString {
-			t.Logf("BlockTemplate (rendered) '%s': '%s'", match, tpl_out["content"])
-			errFilename := filepath.Base(fmt.Sprintf("%s.error", match))
-			err := ioutil.WriteFile(errFilename, []byte(joinedString), 0600)
+			tpl, err := pongo2.FromFile(match)
 			if err != nil {
-				t.Fatalf(err.Error())
+				t.Fatalf("Error on FromFile('%s'): %s", match, err.Error())
 			}
-			t.Logf("get a complete diff with command: 'diff -ya %s %s'", testFilename, errFilename)
-			t.Errorf("Failed: test_out != tpl_out for %s", match)
-		}
+
+			testFilename := fmt.Sprintf("%s.out", match)
+			testOut, rerr := ioutil.ReadFile(testFilename)
+			if rerr != nil {
+				t.Fatalf("Error on ReadFile('%s'): %s", testFilename, rerr.Error())
+			}
+			tpl_out, err := tpl.ExecuteBlocks(tplContext, []string{"content", "more_content"})
+			if err != nil {
+				t.Fatalf("Error on ExecuteBlocks('%s'): %s", match, err.Error())
+			}
+
+			if _, ok := tpl_out["content"]; !ok {
+				t.Errorf("Failed: content not in tpl_out for %s", match)
+			}
+			if _, ok := tpl_out["more_content"]; !ok {
+				t.Errorf("Failed: more_content not in tpl_out for %s", match)
+			}
+			testString := string(testOut[:])
+			joinedString := strings.Join([]string{tpl_out["content"], tpl_out["more_content"]}, "")
+			if testString != joinedString {
+				t.Logf("BlockTemplate (rendered) '%s': '%s'", match, tpl_out["content"])
+				errFilename := filepath.Base(fmt.Sprintf("%s.error", match))
+				err := ioutil.WriteFile(errFilename, []byte(joinedString), 0o600)
+				if err != nil {
+					t.Fatalf(err.Error())
+				}
+				t.Logf("get a complete diff with command: 'diff -ya %s %s'", testFilename, errFilename)
+				t.Errorf("Failed: test_out != tpl_out for %s", match)
+			}
+		})
 	}
 }
 
@@ -440,7 +476,7 @@ var testTemplateFixes = testTemplateFixesT{
 }
 
 func TestExecutionErrors(t *testing.T) {
-	//debug = true
+	// debug = true
 
 	matches, err := filepath.Glob("./template_tests/*-execution.err")
 	if err != nil {
@@ -449,6 +485,9 @@ func TestExecutionErrors(t *testing.T) {
 	for idx, match := range matches {
 		t.Run(fmt.Sprintf("%03d-%s", idx+1, match), func(t *testing.T) {
 			testData, err := ioutil.ReadFile(match)
+			if err != nil {
+				t.Fatalf("could not read file '%v': %v", match, err)
+			}
 			tests := strings.Split(string(testData), "\n")
 
 			checkFilename := fmt.Sprintf("%s.out", match)
@@ -471,12 +510,12 @@ func TestExecutionErrors(t *testing.T) {
 						match, idx+1)
 				}
 
-				tpl, err := pongo2.FromString(test)
+				_, err = pongo2.FromString(test)
 				if err != nil {
 					t.Fatalf("Error on FromString('%s'): %s", test, err.Error())
 				}
 
-				tpl, err = pongo2.FromBytes([]byte(test))
+				tpl, err := pongo2.FromBytes([]byte(test))
 				if err != nil {
 					t.Fatalf("Error on FromBytes('%s'): %s", test, err.Error())
 				}
@@ -498,7 +537,7 @@ func TestExecutionErrors(t *testing.T) {
 }
 
 func TestCompilationErrors(t *testing.T) {
-	//debug = true
+	// debug = true
 
 	matches, err := filepath.Glob("./template_tests/*-compilation.err")
 	if err != nil {
@@ -506,14 +545,16 @@ func TestCompilationErrors(t *testing.T) {
 	}
 	for idx, match := range matches {
 		t.Run(fmt.Sprintf("%03d-%s", idx+1, match), func(t *testing.T) {
-
 			testData, err := ioutil.ReadFile(match)
+			if err != nil {
+				t.Fatalf("could not read file '%v': %v", match, err)
+			}
 			tests := strings.Split(string(testData), "\n")
 
 			checkFilename := fmt.Sprintf("%s.out", match)
 			checkData, err := ioutil.ReadFile(checkFilename)
 			if err != nil {
-				t.Fatalf("Error on ReadFile('%s'): %s", checkFilename, err.Error())
+				t.Fatalf("error on ReadFile('%s'): %s", checkFilename, err.Error())
 			}
 			checks := strings.Split(string(checkData), "\n")
 
@@ -724,6 +765,36 @@ func BenchmarkExecuteBlocksWithSandboxActive(b *testing.B) {
 	}
 }
 
+func BenchmarkExecuteBlocksDeepWithSandboxActive(b *testing.B) {
+	blockNames := []string{"body", "more_content"}
+	tpl, err := pongo2.FromFile("template_tests/block_render/deep.tpl")
+	if err != nil {
+		b.Fatal(err)
+	}
+	b.ResetTimer()
+	for i := 0; i < b.N; i++ {
+		_, err = tpl.ExecuteBlocks(tplContext, blockNames)
+		if err != nil {
+			b.Fatal(err)
+		}
+	}
+}
+
+func BenchmarkExecuteBlocksWithEmptyBlocksSandboxActive(b *testing.B) {
+	blockNames := []string{}
+	tpl, err := pongo2.FromFile("template_tests/block_render/block.tpl")
+	if err != nil {
+		b.Fatal(err)
+	}
+	b.ResetTimer()
+	for i := 0; i < b.N; i++ {
+		_, err = tpl.ExecuteBlocks(tplContext, blockNames)
+		if err != nil {
+			b.Fatal(err)
+		}
+	}
+}
+
 func BenchmarkExecuteBlocksWithoutSandbox(b *testing.B) {
 	blockNames := []string{"content", "more_content"}
 	s := pongo2.NewSet("set without sandbox", pongo2.MustNewLocalFileSystemLoader(""))
